package java.time;

import java.util.concurrent.ConcurrentHashMap;
import java.io.DataInput;
import java.io.IOException;
import java.io.DataOutput;
import java.io.InvalidObjectException;
import java.io.ObjectInputStream;
import java.time.temporal.Temporal;
import java.time.temporal.TemporalQuery;
import java.time.temporal.UnsupportedTemporalTypeException;
import java.time.temporal.ValueRange;
import java.time.temporal.ChronoField;
import java.time.temporal.TemporalField;
import java.time.zone.ZoneRules;
import java.time.temporal.TemporalQueries;
import java.util.Objects;
import java.util.concurrent.ConcurrentMap;
import java.io.Serializable;
import java.time.temporal.TemporalAdjuster;
import java.time.temporal.TemporalAccessor;

public final class ZoneOffset extends ZoneId implements TemporalAccessor, TemporalAdjuster, Comparable<ZoneOffset>, Serializable
{
    private static final ConcurrentMap<Integer, ZoneOffset> SECONDS_CACHE;
    private static final ConcurrentMap<String, ZoneOffset> ID_CACHE;
    private static final int MAX_SECONDS = 64800;
    private static final long serialVersionUID = 2357656521762053153L;
    public static final ZoneOffset UTC;
    public static final ZoneOffset MIN;
    public static final ZoneOffset MAX;
    private final int totalSeconds;
    private final transient String id;
    
    public static ZoneOffset of(String string) {
        Objects.requireNonNull(string, "offsetId");
        ZoneOffset zoneOffset = (ZoneOffset)ZoneOffset.ID_CACHE.get(string);
        if (zoneOffset != null) {
            return zoneOffset;
        }
        int n;
        int n2;
        int n3;
        switch (string.length()) {
            case 2: {
                string = string.charAt(0) + "0" + string.charAt(1);
            }
            case 3: {
                n = parseNumber(string, 1, false);
                n2 = 0;
                n3 = 0;
                break;
            }
            case 5: {
                n = parseNumber(string, 1, false);
                n2 = parseNumber(string, 3, false);
                n3 = 0;
                break;
            }
            case 6: {
                n = parseNumber(string, 1, false);
                n2 = parseNumber(string, 4, true);
                n3 = 0;
                break;
            }
            case 7: {
                n = parseNumber(string, 1, false);
                n2 = parseNumber(string, 3, false);
                n3 = parseNumber(string, 5, false);
                break;
            }
            case 9: {
                n = parseNumber(string, 1, false);
                n2 = parseNumber(string, 4, true);
                n3 = parseNumber(string, 7, true);
                break;
            }
            default: {
                throw new DateTimeException("Invalid ID for ZoneOffset, invalid format: " + string);
            }
        }
        int char1 = string.charAt(0);
        if (char1 != '+' && char1 != '-') {
            throw new DateTimeException("Invalid ID for ZoneOffset, plus/minus not found when expected: " + string);
        }
        if (char1 == '-') {
            return ofHoursMinutesSeconds(-n, -n2, -n3);
        }
        return ofHoursMinutesSeconds(n, n2, n3);
    }
    
    private static int parseNumber(CharSequence charSequence, int n, boolean b) {
        if (b && charSequence.charAt(n - 1) != ':') {
            throw new DateTimeException("Invalid ID for ZoneOffset, colon not found when expected: " + charSequence);
        }
        int char1 = charSequence.charAt(n);
        int char2 = charSequence.charAt(n + 1);
        if (char1 < '0' || char1 > '9' || char2 < '0' || char2 > '9') {
            throw new DateTimeException("Invalid ID for ZoneOffset, non numeric characters found: " + charSequence);
        }
        return (char1 - 48) * 10 + (char2 - 48);
    }
    
    public static ZoneOffset ofHours(int n) {
        return ofHoursMinutesSeconds(n, 0, 0);
    }
    
    public static ZoneOffset ofHoursMinutes(int n, int n2) {
        return ofHoursMinutesSeconds(n, n2, 0);
    }
    
    public static ZoneOffset ofHoursMinutesSeconds(int n, int n2, int n3) {
        validate(n, n2, n3);
        return ofTotalSeconds(totalSeconds(n, n2, n3));
    }
    
    public static ZoneOffset from(TemporalAccessor temporalAccessor) {
        Objects.requireNonNull(temporalAccessor, "temporal");
        ZoneOffset zoneOffset = (ZoneOffset)temporalAccessor.query(TemporalQueries.offset());
        if (zoneOffset == null) {
            throw new DateTimeException("Unable to obtain ZoneOffset from TemporalAccessor: " + temporalAccessor + " of type " + temporalAccessor.getClass().getName());
        }
        return zoneOffset;
    }
    
    private static void validate(int n, int i, int j) {
        if (n < -18 || n > 18) {
            throw new DateTimeException("Zone offset hours not in valid range: value " + n + " is not in the range -18 to 18");
        }
        if (n > 0) {
            if (i < 0 || j < 0) {
                throw new DateTimeException("Zone offset minutes and seconds must be positive because hours is positive");
            }
        }
        else if (n < 0) {
            if (i > 0 || j > 0) {
                throw new DateTimeException("Zone offset minutes and seconds must be negative because hours is negative");
            }
        }
        else if ((i > 0 && j < 0) || (i < 0 && j > 0)) {
            throw new DateTimeException("Zone offset minutes and seconds must have the same sign");
        }
        if (i < -59 || i > 59) {
            throw new DateTimeException("Zone offset minutes not in valid range: value " + i + " is not in the range -59 to 59");
        }
        if (j < -59 || j > 59) {
            throw new DateTimeException("Zone offset seconds not in valid range: value " + j + " is not in the range -59 to 59");
        }
        if (Math.abs(n) == 18 && (i | j) != 0x0) {
            throw new DateTimeException("Zone offset not in valid range: -18:00 to +18:00");
        }
        return;
    }
    
    private static int totalSeconds(int n, int n2, int n3) {
        return n * 3600 + n2 * 60 + n3;
    }
    
    public static ZoneOffset ofTotalSeconds(int i) {
        if (i < -64800 || i > 64800) {
            throw new DateTimeException("Zone offset not in valid range: -18:00 to +18:00");
        }
        if (i % 900 == 0) {
            Integer value = Integer.valueOf(i);
            ZoneOffset zoneOffset = (ZoneOffset)ZoneOffset.SECONDS_CACHE.get(value);
            if (zoneOffset == null) {
                ZoneOffset.SECONDS_CACHE.putIfAbsent(value, new ZoneOffset(i));
                zoneOffset = (ZoneOffset)ZoneOffset.SECONDS_CACHE.get(value);
                ZoneOffset.ID_CACHE.putIfAbsent(zoneOffset.getId(), zoneOffset);
                return zoneOffset;
            }
            return zoneOffset;
        }
        else {
            return new ZoneOffset(i);
        }
    }
    
    private ZoneOffset(int totalSeconds) {
        this.totalSeconds = totalSeconds;
        this.id = buildId(totalSeconds);
    }
    
    private static String buildId(int a) {
        if (a == 0) {
            return "Z";
        }
        int abs = Math.abs(a);
        StringBuilder sb = new StringBuilder();
        int i = abs / 3600;
        int j = abs / 60 % 60;
        sb.append((a < 0) ? "-" : "+").append((i < 10) ? "0" : "").append(i).append((j < 10) ? ":0" : ":").append(j);
        int k = abs % 60;
        if (k != 0) {
            sb.append((k < 10) ? ":0" : ":").append(k);
        }
        return sb.toString();
    }
    
    public int getTotalSeconds() {
        return this.totalSeconds;
    }
    
    @Override
    public String getId() {
        return this.id;
    }
    
    @Override
    public ZoneRules getRules() {
        return ZoneRules.of(this);
    }
    
    @Override
    public boolean isSupported(TemporalField temporalField) {
        if ((temporalField instanceof ChronoField)) {
            return temporalField == ChronoField.OFFSET_SECONDS;
        }
        return temporalField != null && temporalField.isSupportedBy(this);
    }
    
    @Override
    public ValueRange range(TemporalField field) {
        return super.range(field);
    }
    
    @Override
    public int get(TemporalField temporalField) {
        if (temporalField == ChronoField.OFFSET_SECONDS) {
            return this.totalSeconds;
        }
        if ((temporalField instanceof ChronoField)) {
            throw new UnsupportedTemporalTypeException("Unsupported field: " + temporalField);
        }
        return this.range(temporalField).checkValidIntValue(this.getLong(temporalField), temporalField);
    }
    
    @Override
    public long getLong(TemporalField obj) {
        if (obj == ChronoField.OFFSET_SECONDS) {
            return (long)this.totalSeconds;
        }
        if ((obj instanceof ChronoField)) {
            throw new UnsupportedTemporalTypeException("Unsupported field: " + obj);
        }
        return obj.getFrom(this);
    }
    
    @Override
    public <R> R query(TemporalQuery<R> query) {
        if (query == TemporalQueries.offset() || query == TemporalQueries.zone()) {
            return this;
        }
        return super.query(query);
    }
    
    @Override
    public Temporal adjustInto(Temporal temporal) {
        return temporal.with(ChronoField.OFFSET_SECONDS, (long)this.totalSeconds);
    }
    
    @Override
    public int compareTo(ZoneOffset zoneOffset) {
        return zoneOffset.totalSeconds - this.totalSeconds;
    }
    
    @Override
    public boolean equals(Object o) {
        return this == o || (o instanceof ZoneOffset && this.totalSeconds == ((ZoneOffset)o).totalSeconds);
    }
    
    @Override
    public int hashCode() {
        return this.totalSeconds;
    }
    
    @Override
    public String toString() {
        return this.id;
    }
    
    private Object writeReplace() {
        return new Ser((byte)8, this);
    }
    
    private void readObject(ObjectInputStream objectInputStream) throws InvalidObjectException {
        throw new InvalidObjectException("Deserialization via serialization delegate");
    }
    
    @Override
    void write(DataOutput dataOutput) throws IOException {
        dataOutput.writeByte(8);
        this.writeExternal(dataOutput);
    }
    
    void writeExternal(DataOutput dataOutput) throws IOException {
        int totalSeconds = this.totalSeconds;
        int n = (totalSeconds % 900 == 0) ? (totalSeconds / 900) : 127;
        dataOutput.writeByte(n);
        if (n == 127) {
            dataOutput.writeInt(totalSeconds);
        }
    }
    
    static ZoneOffset readExternal(DataInput dataInput) throws IOException {
        int byte1 = dataInput.readByte();
        return (byte1 == 127) ? ofTotalSeconds(dataInput.readInt()) : ofTotalSeconds(byte1 * 900);
    }
    
    static {
        SECONDS_CACHE = new ConcurrentHashMap(16, 0.75f, 4);
        ID_CACHE = new ConcurrentHashMap(16, 0.75f, 4);
        UTC = ofTotalSeconds(0);
        MIN = ofTotalSeconds(-64800);
        MAX = ofTotalSeconds(64800);
    }
}
